package nucleus

import (
	"reflect"
	"sync"

	"code.fbi.h-da.de/cocsn/gosdn/nucleus/errors"

	"github.com/google/uuid"
	log "github.com/sirupsen/logrus"
)

var storeLock sync.RWMutex

// Storable provides an interface for the controller's storage architecture.
type Storable interface {
	ID() uuid.UUID
}

// Store describes an interface for store implementations.
type Store interface {
	Exists(id uuid.UUID) bool
	Add(item Storable) error
	Get(id uuid.UUID) (Storable, error)
	Delete(id uuid.UUID) error
	UUIDs() []uuid.UUID
}

// NewStore returns a generic Store
func NewStore() Store {
	return store{}
}

// NewPndStore returns a PndStore
func NewPndStore() *PndStore {
	return &PndStore{store{}}
}

// NewSbiStore returns a SbiStore
func NewSbiStore() *SbiStore {
	return &SbiStore{store{}}
}

// NewDeviceStore returns a DeviceStore
func NewDeviceStore() *DeviceStore {
	return &DeviceStore{store{}}
}

// NewChangeStore returns a ChangeStore
func NewChangeStore() *ChangeStore {
	return &ChangeStore{store{}}
}

type store map[uuid.UUID]Storable

// Exists takes a Storable's UUID and checks its existence in the store.
func (s store) Exists(id uuid.UUID) bool {
	storeLock.RLock()
	defer storeLock.RUnlock()
	_, ok := s[id]
	return ok
}

// Add adds a Storable to the Store
func (s store) Add(item Storable) error {
	if s.Exists(item.ID()) {
		return &errors.ErrAlreadyExists{Item: item}
	}
	storeLock.Lock()
	s[item.ID()] = item
	storeLock.Unlock()
	log.WithFields(log.Fields{
		"type": reflect.TypeOf(item),
		"uuid": item.ID(),
	}).Debug("storable was added")
	return nil
}

// Get takes a Storable's UUID and returns the Storable. If the requested
// Storable does not exist an error is returned.
func (s store) Get(id uuid.UUID) (Storable, error) {
	if !s.Exists(id) {
		return nil, &errors.ErrNotFound{ID: id}
	}
	log.WithFields(log.Fields{
		"uuid": id,
	}).Debug("storable was accessed")
	storeLock.RLock()
	defer storeLock.RUnlock()
	return s[id], nil
}

// Delete takes a Storable's UUID and deletes it. If the specified UUID does not
// exist in the Store an error is returned.
func (s store) Delete(id uuid.UUID) error {
	if !s.Exists(id) {
		return &errors.ErrNotFound{ID: id}
	}
	storeLock.Lock()
	delete(s, id)
	storeLock.Unlock()
	log.WithFields(log.Fields{
		"uuid": id,
	}).Debug("storable was deleted")
	return nil
}

// UUIDs returns all UUIDs in the store.
func (s store) UUIDs() []uuid.UUID {
	storeLock.RLock()
	defer storeLock.RUnlock()
	keys := make([]uuid.UUID, len(s))
	i := 0
	for k := range s {
		keys[i] = k
		i++
	}
	return keys
}

// SbiStore is used to store SouthboundInterfaces
type SbiStore struct {
	store
}

// Get takes a SouthboundInterface's UUID and returns the SouthboundInterface. If the requested
// SouthboundInterface does not exist an error is returned.
func (s SbiStore) Get(id uuid.UUID) (SouthboundInterface, error) {
	item, err := s.store.Get(id)
	if err != nil {
		return nil, err
	}
	sbi, ok := item.(SouthboundInterface)
	if !ok {
		return nil, &errors.ErrInvalidTypeAssertion{
			Value: sbi,
			Type:  "SouthboundInterface",
		}
	}
	log.WithFields(log.Fields{
		"uuid": id,
	}).Debug("southbound interface was accessed")
	return sbi, nil
}

// PndStore is used to store PrincipalNetworkDomains
type PndStore struct {
	store
}

// Get takes a PrincipalNetworkDomain's UUID and returns the PrincipalNetworkDomain. If the requested
// PrincipalNetworkDomain does not exist an error is returned.
func (s PndStore) Get(id uuid.UUID) (PrincipalNetworkDomain, error) {
	item, err := s.store.Get(id)
	if err != nil {
		return nil, err
	}
	pnd, ok := item.(PrincipalNetworkDomain)
	if !ok {
		return nil, &errors.ErrInvalidTypeAssertion{
			Value: pnd,
			Type:  "PrincipalNetworkDomain",
		}
	}
	log.WithFields(log.Fields{
		"uuid": id,
	}).Debug("principal network domain was accessed")
	return pnd, nil
}

// DeviceStore is used to store Devices
type DeviceStore struct {
	store
}

// Get takes a Device's UUID and returns the Device. If the requested
// Device does not exist an error is returned.
func (s DeviceStore) Get(id uuid.UUID) (*Device, error) {
	item, err := s.store.Get(id)
	if err != nil {
		return nil, err
	}
	device, ok := item.(*Device)
	if !ok {
		return nil, &errors.ErrInvalidTypeAssertion{
			Value: device,
			Type:  reflect.TypeOf(&Device{}),
		}
	}
	log.WithFields(log.Fields{
		"uuid": id,
	}).Debug("device was accessed")
	return device, nil
}

// ChangeStore is used to store Changes
type ChangeStore struct {
	store
}

// Get takes a Change's UUID and returns the Change. If the requested
// Change does not exist an error is returned.
func (s ChangeStore) Get(id uuid.UUID) (*Change, error) {
	item, err := s.store.Get(id)
	if err != nil {
		return nil, err
	}
	change, ok := item.(*Change)
	if !ok {
		return nil, &errors.ErrInvalidTypeAssertion{
			Value: change,
			Type:  reflect.TypeOf(&Change{}),
		}
	}
	log.WithFields(log.Fields{
		"uuid": id,
	}).Debug("change was accessed")
	return change, nil
}
